@@ -83,12 +83,13 @@ typedef char * ptr_t;	/* A generic pointer to which we can add	*/
 #   define VOLATILE
 #endif
 
-#if 0 /* defined(__GNUC__) doesn't work yet */
+#if defined(__GNUC__) && (__GNUC__ > 2) && defined(__OPTIMIZE__)
+/* This doesn't work in some earlier gcc versions */
 # define EXPECT(expr, outcome) __builtin_expect(expr,outcome)
   /* Equivalent to (expr), but predict that usually (expr)==outcome. */
 #else
 # define EXPECT(expr, outcome) (expr)
-#endif /* __GNUC__ */
+#endif
 
 # ifndef GC_LOCKS_H
 #   include "gc_locks.h"
@@ -764,7 +765,7 @@ struct hblk {
 	    /* Under NT, we add only written pages, which can result 	*/
 	    /* in many small root sets.					*/
 #     else
-#       define MAX_ROOT_SETS 256
+#       define MAX_ROOT_SETS 1024
 #     endif
 #   endif
 # endif
