@@ -19,11 +19,13 @@
 #include "cil-coff.h"
 #include "rawbuffer.h"
 #include "mono-endian.h"
-#include "private.h"
 #include "tabledefs.h"
 #include "tokentype.h"
 #include "metadata-internals.h"
 #include <mono/io-layer/io-layer.h>
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <unistd.h>
 
 #define INVALID_ADDRESS 0xffffffff
 
@@ -53,8 +55,9 @@ mono_cli_rva_image_map (MonoCLIImageInfo *iinfo, guint32 addr)
 }
 
 char *
-mono_cli_rva_map (MonoCLIImageInfo *iinfo, guint32 addr)
+mono_image_rva_map (MonoImage *image, guint32 addr)
 {
+	MonoCLIImageInfo *iinfo = image->image_info;
 	const int top = iinfo->cli_section_count;
 	MonoSectionTable *tables = iinfo->cli_section_tables;
 	int i;
@@ -62,6 +65,10 @@ mono_cli_rva_map (MonoCLIImageInfo *iinfo, guint32 addr)
 	for (i = 0; i < top; i++){
 		if ((addr >= tables->st_virtual_address) &&
 		    (addr < tables->st_virtual_address + tables->st_raw_data_size)){
+			if (!iinfo->cli_sections [i]) {
+				if (!mono_image_ensure_section_idx (image, i))
+					return NULL;
+			}
 			return (char*)iinfo->cli_sections [i] +
 				(addr - tables->st_virtual_address);
 		}
@@ -70,6 +77,48 @@ mono_cli_rva_map (MonoCLIImageInfo *iinfo, guint32 addr)
 	return NULL;
 }
 
+static gchar *
+canonicalize_path (const char *path)
+{
+	gchar *abspath, *pos, *lastpos, *dest;
+	int backc;
+
+	if (g_path_is_absolute (path)) {
+		abspath = g_strdup (path);
+	} else {
+		gchar *tmpdir = g_get_current_dir ();
+		abspath = g_build_filename (tmpdir, path, NULL);
+		g_free (tmpdir);
+	}
+
+	abspath = g_strreverse (abspath);
+
+	backc = 0;
+	dest = lastpos = abspath;
+	pos = strchr (lastpos, G_DIR_SEPARATOR);
+
+	while (pos != NULL) {
+		int len = pos - lastpos;
+		if (len == 1 && lastpos [0] == '.') {
+			// nop
+		} else if (len == 2 && lastpos [0] == '.' && lastpos [1] == '.') {
+			backc++;
+		} else if (len > 0) {
+			if (backc > 0) {
+				backc--;
+			} else {
+				if (dest != lastpos) strncpy (dest, lastpos, len + 1);
+				dest += len + 1;
+			}
+		}
+		lastpos = pos + 1;
+		pos = strchr (lastpos, G_DIR_SEPARATOR);
+	}
+	
+	if (dest != lastpos) strcpy (dest, lastpos);
+	return g_strreverse (abspath);
+}
+
 /**
  * mono_images_init:
  *
@@ -110,20 +159,10 @@ mono_image_ensure_section_idx (MonoImage *image, int section)
 	
 	writable = sect->st_flags & SECT_FLAGS_MEM_WRITE;
 
-	if (!image->f) {
-		if (sect->st_raw_data_ptr + sect->st_raw_data_size > image->raw_data_len)
-			return FALSE;
-		/* FIXME: we ignore the writable flag since we don't patch the binary */
-		iinfo->cli_sections [section] = image->raw_data + sect->st_raw_data_ptr;
-		return TRUE;
-	}
-	iinfo->cli_sections [section] = mono_raw_buffer_load (
-		fileno (image->f), writable,
-		sect->st_raw_data_ptr, sect->st_raw_data_size);
-
-	if (iinfo->cli_sections [section] == NULL)
+	if (sect->st_raw_data_ptr + sect->st_raw_data_size > image->raw_data_len)
 		return FALSE;
-
+	/* FIXME: we ignore the writable flag since we don't patch the binary */
+	iinfo->cli_sections [section] = image->raw_data + sect->st_raw_data_ptr;
 	return TRUE;
 }
 
@@ -165,15 +204,10 @@ load_section_tables (MonoImage *image, MonoCLIImageInfo *iinfo, guint32 offset)
 	for (i = 0; i < top; i++){
 		MonoSectionTable *t = &iinfo->cli_section_tables [i];
 
-		if (!image->f) {
-			if (offset + sizeof (MonoSectionTable) > image->raw_data_len)
-				return FALSE;
-			memcpy (t, image->raw_data + offset, sizeof (MonoSectionTable));
-			offset += sizeof (MonoSectionTable);
-		} else {
-			if (fread (t, sizeof (MonoSectionTable), 1, image->f) != 1)
-				return FALSE;
-		}
+		if (offset + sizeof (MonoSectionTable) > image->raw_data_len)
+			return FALSE;
+		memcpy (t, image->raw_data + offset, sizeof (MonoSectionTable));
+		offset += sizeof (MonoSectionTable);
 
 #if G_BYTE_ORDER != G_LITTLE_ENDIAN
 		t->st_virtual_size = GUINT32_FROM_LE (t->st_virtual_size);
@@ -189,10 +223,6 @@ load_section_tables (MonoImage *image, MonoCLIImageInfo *iinfo, guint32 offset)
 		/* consistency checks here */
 	}
 
-	for (i = 0; i < top; i++)
-		if (!mono_image_ensure_section_idx (image, i))
-			return FALSE;
-	
 	return TRUE;
 }
 
@@ -200,23 +230,14 @@ static gboolean
 load_cli_header (MonoImage *image, MonoCLIImageInfo *iinfo)
 {
 	guint32 offset;
-	int n;
 	
 	offset = mono_cli_rva_image_map (iinfo, iinfo->cli_header.datadir.pe_cli_header.rva);
 	if (offset == INVALID_ADDRESS)
 		return FALSE;
 
-	if (!image->f) {
-		if (offset + sizeof (MonoCLIHeader) > image->raw_data_len)
-			return FALSE;
-		memcpy (&iinfo->cli_cli_header, image->raw_data + offset, sizeof (MonoCLIHeader));
-	} else {
-		if (fseek (image->f, offset, SEEK_SET) != 0)
-			return FALSE;
-	
-		if ((n = fread (&iinfo->cli_cli_header, sizeof (MonoCLIHeader), 1, image->f)) != 1)
-			return FALSE;
-	}
+	if (offset + sizeof (MonoCLIHeader) > image->raw_data_len)
+		return FALSE;
+	memcpy (&iinfo->cli_cli_header, image->raw_data + offset, sizeof (MonoCLIHeader));
 
 #if G_BYTE_ORDER != G_LITTLE_ENDIAN
 #define SWAP32(x) (x) = GUINT32_FROM_LE ((x))
@@ -284,15 +305,9 @@ load_metadata_ptrs (MonoImage *image, MonoCLIImageInfo *iinfo)
 
 	size = iinfo->cli_cli_header.ch_metadata.size;
 
-	if (!image->f) {
-		if (offset + size > image->raw_data_len)
-			return FALSE;
-		image->raw_metadata = image->raw_data + offset;
-	} else {
-		image->raw_metadata = mono_raw_buffer_load (fileno (image->f), FALSE, offset, size);
-		if (image->raw_metadata == NULL)
-			return FALSE;
-	}
+	if (offset + size > image->raw_data_len)
+		return FALSE;
+	image->raw_metadata = image->raw_data + offset;
 
 	ptr = image->raw_metadata;
 
@@ -380,16 +395,19 @@ load_tables (MonoImage *image)
 	
 	for (table = 0; table < 64; table++){
 		if ((valid_mask & ((guint64) 1 << table)) == 0){
+			if (table > MONO_TABLE_LAST)
+				continue;
 			image->tables [table].rows = 0;
 			continue;
 		}
 		if (table > MONO_TABLE_LAST) {
 			g_warning("bits in valid must be zero above 0x2d (II - 23.1.6)");
+		} else {
+			image->tables [table].rows = read32 (rows);
 		}
 		/*if ((sorted_mask & ((guint64) 1 << table)) == 0){
 			g_print ("table %s (0x%02x) is sorted\n", mono_meta_table_name (table), table);
 		}*/
-		image->tables [table].rows = read32 (rows);
 		rows++;
 		valid++;
 	}
@@ -438,6 +456,7 @@ load_modules (MonoImage *image, MonoImageOpenStatus *status)
 
 	t = &image->tables [MONO_TABLE_MODULEREF];
 	image->modules = g_new0 (MonoImage *, t->rows);
+	image->module_count = t->rows;
 	base_dir = g_path_get_dirname (image->name);
 	for (i = 0; i < t->rows; i++){
 		char *module_ref;
@@ -540,11 +559,12 @@ build_guid_table (void)
 void
 mono_image_init (MonoImage *image)
 {
+	image->mempool = mono_mempool_new ();
 	image->method_cache = g_hash_table_new (NULL, NULL);
 	image->class_cache = g_hash_table_new (NULL, NULL);
 	image->field_cache = g_hash_table_new (NULL, NULL);
 	image->name_cache = g_hash_table_new (g_str_hash, g_str_equal);
-	image->array_cache = g_hash_table_new (NULL, NULL);
+	image->array_cache = g_hash_table_new (mono_aligned_addr_hash, NULL);
 
 	image->delegate_begin_invoke_cache = 
 		g_hash_table_new ((GHashFunc)mono_signature_hash, 
@@ -559,27 +579,24 @@ mono_image_init (MonoImage *image)
 		g_hash_table_new ((GHashFunc)mono_signature_hash, 
 				  (GCompareFunc)mono_metadata_signature_equal);
 	
-	image->managed_wrapper_cache = g_hash_table_new (NULL, NULL);
-	image->native_wrapper_cache = g_hash_table_new (NULL, NULL);
-	image->remoting_invoke_cache = g_hash_table_new (NULL, NULL);
-	image->synchronized_cache = g_hash_table_new (NULL, NULL);
-	image->unbox_wrapper_cache = g_hash_table_new (NULL, NULL);
+	image->managed_wrapper_cache = g_hash_table_new (mono_aligned_addr_hash, NULL);
+	image->native_wrapper_cache = g_hash_table_new (mono_aligned_addr_hash, NULL);
+	image->remoting_invoke_cache = g_hash_table_new (mono_aligned_addr_hash, NULL);
+	image->synchronized_cache = g_hash_table_new (mono_aligned_addr_hash, NULL);
+	image->unbox_wrapper_cache = g_hash_table_new (mono_aligned_addr_hash, NULL);
 
 	image->typespec_cache = g_hash_table_new (NULL, NULL);
 	image->memberref_signatures = g_hash_table_new (NULL, NULL);
-
-	image->generic_inst_cache =
-		g_hash_table_new ((GHashFunc)mono_metadata_generic_inst_hash,
-				  (GCompareFunc)mono_metadata_generic_inst_equal);
+	image->helper_signatures = g_hash_table_new (g_str_hash, g_str_equal);
 }
 
 static MonoImage *
-do_mono_image_load (MonoImage *image, MonoImageOpenStatus *status)
+do_mono_image_load (MonoImage *image, MonoImageOpenStatus *status,
+		    gboolean care_about_cli)
 {
 	MonoCLIImageInfo *iinfo;
 	MonoDotNetHeader *header;
 	MonoMSDOSHeader msdos;
-	int n;
 	guint32 offset = 0;
 
 	mono_image_init (image);
@@ -590,35 +607,21 @@ do_mono_image_load (MonoImage *image, MonoImageOpenStatus *status)
 	if (status)
 		*status = MONO_IMAGE_IMAGE_INVALID;
 
-	if (!image->f) {
-		if (offset + sizeof (msdos) > image->raw_data_len)
-			goto invalid_image;
-		memcpy (&msdos, image->raw_data + offset, sizeof (msdos));
-	} else {
-		if (fread (&msdos, sizeof (msdos), 1, image->f) != 1)
-			goto invalid_image;
-	}
+	if (offset + sizeof (msdos) > image->raw_data_len)
+		goto invalid_image;
+	memcpy (&msdos, image->raw_data + offset, sizeof (msdos));
 	
 	if (!(msdos.msdos_sig [0] == 'M' && msdos.msdos_sig [1] == 'Z'))
 		goto invalid_image;
 	
 	msdos.pe_offset = GUINT32_FROM_LE (msdos.pe_offset);
 
-	if (msdos.pe_offset != sizeof (msdos)) {
-		if (image->f)
-			fseek (image->f, msdos.pe_offset, SEEK_SET);
-	}
 	offset = msdos.pe_offset;
 
-	if (!image->f) {
-		if (offset + sizeof (MonoDotNetHeader) > image->raw_data_len)
-			goto invalid_image;
-		memcpy (header, image->raw_data + offset, sizeof (MonoDotNetHeader));
-		offset += sizeof (MonoDotNetHeader);
-	} else {
-		if ((n = fread (header, sizeof (MonoDotNetHeader), 1, image->f)) != 1)
-			goto invalid_image;
-	}
+	if (offset + sizeof (MonoDotNetHeader) > image->raw_data_len)
+		goto invalid_image;
+	memcpy (header, image->raw_data + offset, sizeof (MonoDotNetHeader));
+	offset += sizeof (MonoDotNetHeader);
 
 #if G_BYTE_ORDER != G_LITTLE_ENDIAN
 #define SWAP32(x) (x) = GUINT32_FROM_LE ((x))
@@ -718,6 +721,10 @@ do_mono_image_load (MonoImage *image, MonoImageOpenStatus *status)
 	if (!load_section_tables (image, iinfo, offset))
 		goto invalid_image;
 	
+	if (care_about_cli == FALSE) {
+		goto done;
+	}
+	
 	/* Load the CLI header */
 	if (!load_cli_header (image, iinfo))
 		goto invalid_image;
@@ -739,6 +746,7 @@ do_mono_image_load (MonoImage *image, MonoImageOpenStatus *status)
 
 	load_modules (image, status);
 
+done:
 	if (status)
 		*status = MONO_IMAGE_OK;
 
@@ -752,11 +760,13 @@ invalid_image:
 }
 
 static MonoImage *
-do_mono_image_open (const char *fname, MonoImageOpenStatus *status)
+do_mono_image_open (const char *fname, MonoImageOpenStatus *status,
+		    gboolean care_about_cli)
 {
 	MonoCLIImageInfo *iinfo;
 	MonoImage *image;
 	FILE *filed;
+	struct stat stat_buf;
 
 	if ((filed = fopen (fname, "rb")) == NULL){
 		if (status)
@@ -764,21 +774,22 @@ do_mono_image_open (const char *fname, MonoImageOpenStatus *status)
 		return NULL;
 	}
 
+	if (fstat (fileno (filed), &stat_buf)) {
+		fclose (filed);
+		if (status)
+			*status = MONO_IMAGE_ERROR_ERRNO;
+		return NULL;
+	}
 	image = g_new0 (MonoImage, 1);
 	image->ref_count = 1;
-	image->f = filed;
+	image->file_descr = filed;
+	image->raw_data_len = stat_buf.st_size;
+	image->raw_data = mono_raw_buffer_load (fileno (filed), FALSE, 0, stat_buf.st_size);
 	iinfo = g_new0 (MonoCLIImageInfo, 1);
 	image->image_info = iinfo;
+	image->name = canonicalize_path (fname);
 
-	if (g_path_is_absolute (fname))
-		image->name = g_strdup (fname);
-	else {
-		gchar *path = g_get_current_dir ();
-		image->name = g_build_filename (path, fname, NULL);
-		g_free (path);
-	}
-
-	return do_mono_image_load (image, status);
+	return do_mono_image_load (image, status, care_about_cli);
 }
 
 MonoImage *
@@ -803,6 +814,30 @@ mono_image_loaded_by_guid (const char *guid)
 	return res;
 }
 
+static MonoImage *
+register_image (MonoImage *image)
+{
+	MonoImage *image2;
+
+	EnterCriticalSection (&images_mutex);
+	image2 = g_hash_table_lookup (loaded_images_hash, image->name);
+
+	if (image2) {
+		/* Somebody else beat us to it */
+		mono_image_addref (image2);
+		LeaveCriticalSection (&images_mutex);
+		mono_image_close (image);
+		return image2;
+	}
+	g_hash_table_insert (loaded_images_hash, image->name, image);
+	if (image->assembly_name && (g_hash_table_lookup (loaded_images_hash, image->assembly_name) == NULL))
+		g_hash_table_insert (loaded_images_hash, (char *) image->assembly_name, image);	
+	g_hash_table_insert (loaded_images_guid_hash, image->guid, image);
+	LeaveCriticalSection (&images_mutex);
+
+	return image;
+}
+
 MonoImage *
 mono_image_open_from_data (char *data, guint32 data_len, gboolean need_copy, MonoImageOpenStatus *status)
 {
@@ -835,7 +870,11 @@ mono_image_open_from_data (char *data, guint32 data_len, gboolean need_copy, Mon
 	iinfo = g_new0 (MonoCLIImageInfo, 1);
 	image->image_info = iinfo;
 
-	return do_mono_image_load (image, status);
+	image = do_mono_image_load (image, status, TRUE);
+	if (image == NULL)
+		return NULL;
+
+	return register_image (image);
 }
 
 /**
@@ -843,24 +882,18 @@ mono_image_open_from_data (char *data, guint32 data_len, gboolean need_copy, Mon
  * @fname: filename that points to the module we want to open
  * @status: An error condition is returned in this field
  *
- * Retuns: An open image of type %MonoImage or NULL on error.
+ * Returns: An open image of type %MonoImage or NULL on error.
  * if NULL, then check the value of @status for details on the error
  */
 MonoImage *
 mono_image_open (const char *fname, MonoImageOpenStatus *status)
 {
-	MonoImage *image, *image2;
+	MonoImage *image;
 	char *absfname;
 	
 	g_return_val_if_fail (fname != NULL, NULL);
 	
-	if (g_path_is_absolute (fname)) 
-		absfname = (char*)fname;
-	else {
-		gchar *path = g_get_current_dir ();
-		absfname = g_build_filename (path, fname, NULL);
-		g_free (path);
-	}
+	absfname = canonicalize_path (fname);
 
 	/*
 	 * The easiest solution would be to do all the loading inside the mutex,
@@ -870,38 +903,39 @@ mono_image_open (const char *fname, MonoImageOpenStatus *status)
 	 */
 	EnterCriticalSection (&images_mutex);
 	image = g_hash_table_lookup (loaded_images_hash, absfname);
-	
-	if (absfname != fname)
-		g_free (absfname);
+	g_free (absfname);
 	
 	if (image){
-		image->ref_count++;
+		mono_image_addref (image);
 		LeaveCriticalSection (&images_mutex);
 		return image;
 	}
 	LeaveCriticalSection (&images_mutex);
 
-	image = do_mono_image_open (fname, status);
+	image = do_mono_image_open (fname, status, TRUE);
 	if (image == NULL)
 		return NULL;
 
-	EnterCriticalSection (&images_mutex);
-	image2 = g_hash_table_lookup (loaded_images_hash, fname);
-	if (image2) {
-		/* Somebody else beat us to it */
-		image2->ref_count ++;
-		LeaveCriticalSection (&images_mutex);
-		mono_image_close (image);
+	return register_image (image);
+}
 
-		return image2;
-	}
-	g_hash_table_insert (loaded_images_hash, image->name, image);
-	if (image->assembly_name && (g_hash_table_lookup (loaded_images_hash, image->assembly_name) == NULL))
-		g_hash_table_insert (loaded_images_hash, (char *) image->assembly_name, image);	
-	g_hash_table_insert (loaded_images_guid_hash, image->guid, image);
-	LeaveCriticalSection (&images_mutex);
-
-	return image;
+/**
+ * mono_pe_file_open:
+ * @fname: filename that points to the module we want to open
+ * @status: An error condition is returned in this field
+ *
+ * Returns: An open image of type %MonoImage or NULL on error.  if
+ * NULL, then check the value of @status for details on the error.
+ * This variant for mono_image_open DOES NOT SET UP CLI METADATA.
+ * It's just a PE file loader, used for FileVersionInfo.  It also does
+ * not use the image cache.
+ */
+MonoImage *
+mono_pe_file_open (const char *fname, MonoImageOpenStatus *status)
+{
+	g_return_val_if_fail (fname != NULL, NULL);
+	
+	return(do_mono_image_open (fname, status, FALSE));
 }
 
 static void
@@ -916,8 +950,14 @@ free_mr_signatures (gpointer key, gpointer val, gpointer user_data)
 	mono_metadata_free_method_signature ((MonoMethodSignature*)val);
 }
 
+static void
+free_remoting_wrappers (gpointer key, gpointer val, gpointer user_data)
+{
+	g_free (val);
+}
+
 /**
- * mono_image_close:
+ * mono_image_addref:
  * @image: The image file we wish to add a reference to
  *
  *  Increases the reference count of an image.
@@ -928,6 +968,18 @@ mono_image_addref (MonoImage *image)
 	InterlockedIncrement (&image->ref_count);
 }	
 
+void
+mono_dynamic_stream_reset (MonoDynamicStream* stream)
+{
+	stream->alloc_size = stream->index = stream->offset = 0;
+	g_free (stream->data);
+	stream->data = NULL;
+	if (stream->hash) {
+		g_hash_table_destroy (stream->hash);
+		stream->hash = NULL;
+	}
+}
+
 /**
  * mono_image_close:
  * @image: The image file we wish to close
@@ -939,10 +991,13 @@ void
 mono_image_close (MonoImage *image)
 {
 	MonoImage *image2;
+	int i;
 
 	g_return_if_fail (image != NULL);
 
 	EnterCriticalSection (&images_mutex);
+	/*g_print ("destroy image '%s' %p (dynamic: %d) refcount: %d\n", image->name, image, image->dynamic, image->ref_count);*/
+	g_assert (image->ref_count > 0);
 	if (--image->ref_count) {
 		LeaveCriticalSection (&images_mutex);
 		return;
@@ -959,12 +1014,16 @@ mono_image_close (MonoImage *image)
 		g_hash_table_remove (loaded_images_hash, (char *) image->assembly_name);	
 	LeaveCriticalSection (&images_mutex);
 
-	if (image->f)
-		fclose (image->f);
+	if (image->file_descr) {
+		fclose (image->file_descr);
+		image->file_descr = NULL;
+		if (image->raw_data != NULL)
+			mono_raw_buffer_free (image->raw_data);
+	}
+	
 	if (image->raw_data_allocated) {
 		/* image->raw_metadata and cli_sections might lie inside image->raw_data */
 		MonoCLIImageInfo *ii = image->image_info;
-		int i;
 
 		if ((image->raw_metadata > image->raw_data) &&
 			(image->raw_metadata <= (image->raw_data + image->raw_data_len)))
@@ -978,6 +1037,7 @@ mono_image_close (MonoImage *image)
 		g_free (image->raw_data);
 	}
 	g_free (image->name);
+	g_free (image->guid);
 	g_free (image->files);
 
 	g_hash_table_destroy (image->method_cache);
@@ -991,27 +1051,20 @@ mono_image_close (MonoImage *image)
 	g_hash_table_destroy (image->delegate_begin_invoke_cache);
 	g_hash_table_destroy (image->delegate_end_invoke_cache);
 	g_hash_table_destroy (image->delegate_invoke_cache);
+	g_hash_table_foreach (image->remoting_invoke_cache, free_remoting_wrappers, NULL);
 	g_hash_table_destroy (image->remoting_invoke_cache);
 	g_hash_table_destroy (image->runtime_invoke_cache);
 	g_hash_table_destroy (image->synchronized_cache);
 	g_hash_table_destroy (image->unbox_wrapper_cache);
 	g_hash_table_destroy (image->typespec_cache);
-	g_hash_table_destroy (image->generic_inst_cache);
 	g_hash_table_foreach (image->memberref_signatures, free_mr_signatures, NULL);
 	g_hash_table_destroy (image->memberref_signatures);
-	
-	if (image->raw_metadata != NULL)
-		mono_raw_buffer_free (image->raw_metadata);
+	g_hash_table_foreach (image->helper_signatures, free_mr_signatures, NULL);
+	g_hash_table_destroy (image->helper_signatures);
 	
 	if (image->image_info){
 		MonoCLIImageInfo *ii = image->image_info;
-		int i;
 
-		for (i = 0; i < ii->cli_section_count; i++){
-			if (!ii->cli_sections [i])
-				continue;
-			mono_raw_buffer_free (ii->cli_sections [i]);
-		}
 		if (ii->cli_section_tables)
 			g_free (ii->cli_section_tables);
 		if (ii->cli_sections)
@@ -1019,7 +1072,55 @@ mono_image_close (MonoImage *image)
 		g_free (image->image_info);
 	}
 
-	g_free (image);
+	for (i = 0; i < image->module_count; ++i) {
+		if (image->modules [i])
+			mono_image_close (image->modules [i]);
+	}
+	/*g_print ("destroy image %p (dynamic: %d)\n", image, image->dynamic);*/
+	if (!image->dynamic) {
+		mono_mempool_destroy (image->mempool);
+		g_free (image);
+	} else {
+		/* Dynamic images are GC_MALLOCed */
+		struct _MonoDynamicImage *di = (struct _MonoDynamicImage*)image;
+		int i;
+		g_free ((char*)image->module_name);
+		if (di->typespec)
+			g_hash_table_destroy (di->typespec);
+		if (di->typeref)
+			g_hash_table_destroy (di->typeref);
+		if (di->handleref)
+			g_hash_table_destroy (di->handleref);
+		if (di->tokens)
+			mono_g_hash_table_destroy (di->tokens);
+		if (di->blob_cache)
+			g_hash_table_destroy (di->blob_cache);
+		g_list_free (di->array_methods);
+		if (di->gen_params)
+			g_ptr_array_free (di->gen_params, TRUE);
+		if (di->token_fixups)
+			mono_g_hash_table_destroy (di->token_fixups);
+		if (di->method_to_table_idx)
+			g_hash_table_destroy (di->method_to_table_idx);
+		if (di->field_to_table_idx)
+			g_hash_table_destroy (di->field_to_table_idx);
+		if (di->method_aux_hash)
+			g_hash_table_destroy (di->method_aux_hash);
+		g_free (di->strong_name);
+		g_free (di->win32_res);
+		/*g_print ("string heap destroy for image %p\n", di);*/
+		mono_dynamic_stream_reset (&di->sheap);
+		mono_dynamic_stream_reset (&di->code);
+		mono_dynamic_stream_reset (&di->resources);
+		mono_dynamic_stream_reset (&di->us);
+		mono_dynamic_stream_reset (&di->blob);
+		mono_dynamic_stream_reset (&di->tstream);
+		mono_dynamic_stream_reset (&di->guid);
+		for (i = 0; i < MONO_TABLE_NUM; ++i) {
+			g_free (di->tables [i].values);
+		}
+		mono_mempool_destroy (image->mempool);
+	}
 }
 
 /** 
@@ -1076,8 +1177,10 @@ mono_image_walk_resource_tree (MonoCLIImageInfo *info, guint32 res_id,
 		}
 #endif
 	} else if (level==2) {
-		if((is_string==FALSE && entry->name_offset!=lang_id) ||
-		   (is_string==TRUE)) {
+		if ((is_string == FALSE &&
+		    entry->name_offset != lang_id &&
+		    lang_id != 0) ||
+		   (is_string == TRUE)) {
 			return(NULL);
 		}
 	} else {
@@ -1147,7 +1250,7 @@ mono_image_lookup_resource (MonoImage *image, guint32 res_id, guint32 lang_id, g
 		return(NULL);
 	}
 
-	resource_dir=(MonoPEResourceDir *)mono_cli_rva_map (info, rsrc->rva);
+	resource_dir=(MonoPEResourceDir *)mono_image_rva_map (image, rsrc->rva);
 	if(resource_dir==NULL) {
 		return(NULL);
 	}
@@ -1186,7 +1289,7 @@ mono_image_get_resource (MonoImage *image, guint32 offset, guint32 *size)
 	if (!ch->ch_resources.rva || offset + 4 > ch->ch_resources.size)
 		return NULL;
 	
-	data = mono_cli_rva_map (iinfo, ch->ch_resources.rva);
+	data = mono_image_rva_map (image, ch->ch_resources.rva);
 	if (!data)
 		return NULL;
 	data += offset;
@@ -1221,10 +1324,9 @@ mono_image_load_file_for_image (MonoImage *image, int fileidx)
 	res = mono_image_open (name, NULL);
 	if (res) {
 		int i;
-		t = &res->tables [MONO_TABLE_MODULEREF];
 		/* g_print ("loaded file %s from %s (%p)\n", name, image->name, image->assembly); */
 		res->assembly = image->assembly;
-		for (i = 0; i < t->rows; ++i) {
+		for (i = 0; i < res->module_count; ++i) {
 			if (res->modules [i] && !res->modules [i]->assembly)
 				res->modules [i]->assembly = image->assembly;
 		}
@@ -1245,7 +1347,7 @@ mono_image_get_strong_name (MonoImage *image, guint32 *size)
 
 	if (!de->size || !de->rva)
 		return NULL;
-	data = mono_cli_rva_map (iinfo, de->rva);
+	data = mono_image_rva_map (image, de->rva);
 	if (!data)
 		return NULL;
 	if (size)
@@ -1311,7 +1413,7 @@ mono_image_get_filename (MonoImage *image)
 const MonoTableInfo*
 mono_image_get_table_info (MonoImage *image, int table_id)
 {
-	if (table_id < 0 || table_id >= 64)
+	if (table_id < 0 || table_id >= MONO_TABLE_NUM)
 		return NULL;
 	return &image->tables [table_id];
 }
@@ -1319,13 +1421,13 @@ mono_image_get_table_info (MonoImage *image, int table_id)
 int
 mono_image_get_table_rows (MonoImage *image, int table_id)
 {
-	if (table_id < 0 || table_id >= 64)
+	if (table_id < 0 || table_id >= MONO_TABLE_NUM)
 		return 0;
 	return image->tables [table_id].rows;
 }
 
 int
-mono_table_info_get_rows (MonoTableInfo *table)
+mono_table_info_get_rows (const MonoTableInfo *table)
 {
 	return table->rows;
 }
@@ -1342,9 +1444,12 @@ mono_image_is_dynamic (MonoImage *image)
 	return image->dynamic;
 }
 
-char*
-mono_image_rva_map (MonoImage *image, guint32 rva)
+gboolean
+mono_image_has_authenticode_entry (MonoImage *image)
 {
-	return mono_cli_rva_map (image->image_info, rva);
+	MonoCLIImageInfo *iinfo = image->image_info;
+	MonoDotNetHeader *header = &iinfo->cli_header;
+	MonoPEDirEntry *de = &header->datadir.pe_certificate_table;
+	// the Authenticode "pre" (non ASN.1) header is 8 bytes long
+	return ((de->rva != 0) && (de->size > 8));
 }
-
