@@ -3,6 +3,8 @@
 
 #include <mono/metadata/class.h>
 
+G_BEGIN_DECLS
+
 typedef guchar MonoBoolean;
 
 typedef struct _MonoReflectionMethod MonoReflectionMethod;
@@ -18,6 +20,8 @@ typedef struct _MonoThreadsSync MonoThreadsSync;
 typedef struct _MonoThread MonoThread;
 typedef struct _MonoDynamicAssembly MonoDynamicAssembly;
 typedef struct _MonoDynamicImage MonoDynamicImage;
+typedef struct _MonoReflectionMethodBody MonoReflectionMethodBody;
+typedef struct _MonoAppContext MonoAppContext;
 
 typedef struct {
 	MonoVTable *vtable;
@@ -53,7 +57,7 @@ typedef void	    (*MonoMainThreadFunc)    (gpointer user_data);
 #define mono_object_domain(obj) (((MonoObject*)(obj))->vtable->domain)
 
 #define mono_array_length(array) ((array)->max_length)
-#define mono_array_addr(array,type,index) ( ((char*)(array)->vector) + sizeof (type) * (index) )
+#define mono_array_addr(array,type,index) ((type*) mono_array_addr_with_size (array, sizeof (type), index))
 #define mono_array_addr_with_size(array,size,index) ( ((char*)(array)->vector) + (size) * (index) )
 #define mono_array_get(array,type,index) ( *(type*)mono_array_addr ((array), type, (index)) ) 
 #define mono_array_set(array,type,index,value)	\
@@ -127,6 +131,12 @@ mono_string_to_utf16	    (MonoString *string_obj);
 MonoString *
 mono_string_from_utf16	    (gunichar2 *data);
 
+gboolean
+mono_string_equal           (MonoString *s1, MonoString *s2);
+
+guint
+mono_string_hash            (MonoString *s);
+
 MonoObject *
 mono_value_box		    (MonoDomain *domain, MonoClass *klass, gpointer val);
 
@@ -157,6 +167,9 @@ mono_monitor_try_enter       (MonoObject *obj, guint32 ms);
 gboolean
 mono_monitor_enter           (MonoObject *obj);
 
+guint
+mono_object_get_size         (MonoObject *o);
+
 void 
 mono_monitor_exit            (MonoObject *obj);
 
@@ -227,9 +240,6 @@ mono_compile_method	   (MonoMethod *method);
 MonoRemoteClass*
 mono_remote_class (MonoDomain *domain, MonoString *class_name, MonoClass *proxy_class);
 
-void
-mono_upgrade_remote_class (MonoDomain *domain, MonoRemoteClass *remote_class, MonoClass *klass);
-
 /* accessors for fields and properties */
 void
 mono_field_set_value (MonoObject *obj, MonoClassField *field, void *value);
@@ -252,11 +262,26 @@ mono_property_set_value (MonoProperty *prop, void *obj, void **params, MonoObjec
 MonoObject*
 mono_property_get_value (MonoProperty *prop, void *obj, void **params, MonoObject **exc);
 
-/* GC handles support */
+/* GC handles support 
+ *
+ * A handle can be created to refer to a managed object and either prevent it
+ * from being garbage collected or moved or to be able to know if it has been 
+ * collected or not (weak references).
+ * mono_gchandle_new () is used to prevent an object from being garbage collected
+ * until mono_gchandle_free() is called. Use a TRUE value for the pinned argument to
+ * prevent the object from being moved (this should be avoided as much as possible 
+ * and this should be used only for shorts periods of time or performance will suffer).
+ * To create a weakref use mono_gchandle_new_weakref (): track_resurrection should
+ * usually be false (see the GC docs for more details).
+ * mono_gchandle_get_target () can be used to get the object referenced by both kinds
+ * of handle: for a weakref handle, if an object has been collected, it will return NULL.
+ */
 guint32      mono_gchandle_new         (MonoObject *obj, gboolean pinned);
 guint32      mono_gchandle_new_weakref (MonoObject *obj, gboolean track_resurrection);
 MonoObject*  mono_gchandle_get_target  (guint32 gchandle);
 void         mono_gchandle_free        (guint32 gchandle);
 
+G_END_DECLS
+
 #endif
 
