@@ -5,6 +5,8 @@
 #include <mono/metadata/image.h>
 #include <mono/metadata/loader.h>
 
+G_BEGIN_DECLS
+
 typedef enum {
 	MONO_VERIFY_OK,
 	MONO_VERIFY_ERROR,
@@ -23,5 +25,7 @@ GSList* mono_method_verify       (MonoMethod *method, int level);
 void    mono_free_verify_list    (GSList *list);
 char*   mono_verify_corlib       (void);
 
+G_END_DECLS
+
 #endif  /* __MONO_METADATA_VERIFY_H__ */
 
