@@ -11,6 +11,8 @@
  *
  * Taken from the FreeBSD distribution.
  */
+#include <glib.h>
+
 #include "strtod.h"
 
 /*-
@@ -141,20 +143,13 @@ static char sccsid[] = "@(#)strtod.c	8.1 (Berkeley) 6/4/93";
 #if defined(i386) || defined(mips) && defined(MIPSEL) || defined (__arm__)
 
 #define IEEE_8087
-#define Long long
 
 #elif defined(__x86_64__)
 
 #define IEEE_8087
-#define Long int
 
 #elif defined(__ia64)
 
-# ifndef __LP64__
-#  define Long long
-# else
-#  define Long int
-# endif
 # ifdef __hpux
 #  define IEEE_MC68k
 # else
@@ -164,18 +159,13 @@ static char sccsid[] = "@(#)strtod.c	8.1 (Berkeley) 6/4/93";
 #elif defined(__hppa)
 
 # define IEEE_MC68k
-# ifndef __LP64__
-#  define Long long
-# else
-#  define Long int
-# endif
 
 #else
 #define IEEE_MC68k
-#define Long long
 #endif
 
-#define ULong unsigned Long
+#define Long gint32
+#define ULong guint32
 
 #ifdef DEBUG
 #include "stdio.h"
@@ -1200,7 +1190,7 @@ ratio
 	return da / db;
 }
 
- static double
+ static const double
 tens[] = {
 		1e0, 1e1, 1e2, 1e3, 1e4, 1e5, 1e6, 1e7, 1e8, 1e9,
 		1e10, 1e11, 1e12, 1e13, 1e14, 1e15, 1e16, 1e17, 1e18, 1e19,
@@ -1210,19 +1200,19 @@ tens[] = {
 #endif
 		};
 
- static double
+ static const double
 #ifdef IEEE_Arith
 bigtens[] = { 1e16, 1e32, 1e64, 1e128, 1e256 };
-static double tinytens[] = { 1e-16, 1e-32, 1e-64, 1e-128, 1e-256 };
+static const double tinytens[] = { 1e-16, 1e-32, 1e-64, 1e-128, 1e-256 };
 #define n_bigtens 5
 #else
 #ifdef IBM
 bigtens[] = { 1e16, 1e32, 1e64 };
-static double tinytens[] = { 1e-16, 1e-32, 1e-64 };
+static const double tinytens[] = { 1e-16, 1e-32, 1e-64 };
 #define n_bigtens 3
 #else
 bigtens[] = { 1e16, 1e32 };
-static double tinytens[] = { 1e-16, 1e-32 };
+static const double tinytens[] = { 1e-16, 1e-32 };
 #define n_bigtens 2
 #endif
 #endif
